﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.IO;
using System.Linq;
using System.Threading.Tasks;

namespace CSharp_Reference_App
{
   internal class Program
   {
      private const string MyDatasetName = "N8844A C# Reference App Data";
      private static string sUrl;
      private static string sUsername;
      private static string sPassword;
      private static string sOutputPath;
      private static ClientApi sClientApi;
      private static ObservableCollection<Dataset> sDatasetsOnServer;
      private static Dataset sMyDataset;
      private static long sMyDatasetId;

      private static void Main(string[] args)
      {
         ProcessArguments(args);

         try
         {
            var task = Task.Run(() => Execute());
            task.Wait();
         }
         catch (AggregateException e)
         {
            if (e.InnerException == null)
            {
               throw;
            }

            throw e.InnerException;
         }
         finally
         {
            var task = Task.Run(() => LogOut());
            task.Wait();
         }
      }

      private static void ProcessArguments(string[] args)
      {
         if (args.Length == 0)
         {
            sUrl = GetParameterFromUser("Enter URL of N8844A Keysight Data Analytics Web Service (e.g. https://123.45.67.890)");
            sUsername = GetParameterFromUser("Enter username");
            sPassword = GetParameterFromUser("Enter password");
            sOutputPath = GetParameterFromUser("Enter output path (e.g. c:\\temp\\)");
         }
         else if (args.Length == 4)
         {
            sUrl = args[0];
            sUsername = args[1];
            sPassword = args[2];
            sOutputPath = args[3];
         }
         else
         {
            Console.WriteLine("Four arguments expected: URL, username, password, ouput path");
            PressAnyKeyToClose();
         }
      }

      private static string GetParameterFromUser(string prompt)
      {
         Console.WriteLine(prompt);
         return Console.ReadLine();
      }

      private static async Task Execute()
      {
         sClientApi = new ClientApi(sUrl);

         try
         {
            await ValidateServer();
         }
         catch (Exception e)
         {
            Console.WriteLine(
               "Failed to find Keysight N8844A Data Analytics Web Service software running at " + sUrl +
               Environment.NewLine + "Details: " + e.Message);
            PressAnyKeyToClose();
         }

         try
         {
            await LogIn();
         }
         catch (Exception e)
         {
            Console.WriteLine("Failed to login to server at " + sUrl + Environment.NewLine + "Details: " + e.Message);
            PressAnyKeyToClose();
         }

         try
         {
            await CreateDataset();
         }
         catch (Exception e)
         {
            Console.WriteLine("Failed to create dataset '" + MyDatasetName + "'" + Environment.NewLine + "Details: " + e.Message);
            PressAnyKeyToClose();
         }

         try
         {
            await UploadTypesAndMeasurements();
         }
         catch (Exception e)
         {
            Console.WriteLine("Failed to upload types and measurements to server." + Environment.NewLine + "Details: " + e.Message);
            PressAnyKeyToClose();
         }

         try
         {
            await ReportServerContents();
         }
         catch (Exception e)
         {
            Console.WriteLine("Failed to download measurements from server." + Environment.NewLine + "Details: " + e.Message);
            PressAnyKeyToClose();
         }

         PressAnyKeyToClose();
      }

      private static async Task ValidateServer()
      {
         var serverVersion = await sClientApi.GetVersionAsync();
         Console.WriteLine("On server: " + sUrl);
         Console.WriteLine("Found Keysight N8844A Data Analytics Web Service software v" + serverVersion);
      }

      private static async Task LogIn()
      {
         // If your server does not have a signed certificate, enable the following line to ignore the AuthenticationException during login
         // ServicePointManager.ServerCertificateValidationCallback += (sender, cert, chain, sslPolicyErrors) => true;

         // Log in to the server
         await sClientApi.LoginAsync(new UsernamePasswordCredentials {Username = sUsername, Password = sPassword});
         Console.WriteLine("Logged in as " + sUsername);
      }

      private static async Task CreateDataset()
      {
         sDatasetsOnServer = await sClientApi.GetAllDatasetsAsync();
         sMyDataset = sDatasetsOnServer.FirstOrDefault(ds => ds.Name == MyDatasetName);

         if (sMyDataset == null)
         {
            // Dataset I want to use does not exist, so create it
            sMyDataset = await sClientApi.CreateDatasetAsync(new Dataset {Name = MyDatasetName});
            sDatasetsOnServer.Add(sMyDataset);
         }

         if (sMyDataset.Id.HasValue)
         {
            sMyDatasetId = sMyDataset.Id.Value;
         }

         Console.WriteLine("Created dataset '" + MyDatasetName + "'");

         if (!sMyDataset.Id.HasValue)
         {
            Console.WriteLine("Dataset '" + MyDatasetName + " is corrupted");
            PressAnyKeyToClose();
         }
      }

      private static async Task UploadTypesAndMeasurements()
      {
         // Create a measurement type to be uploaded
         var measurementTypeNames = new List<MeasurementType>
         {
            new MeasurementType
            {
               Name = "Frequency",
               LowerLimit = 1E+09,
               UpperLimit = 2E+09,
               Units = "Hz"
            }
         };
         var createdMeasurementTypes = await sClientApi.CreateMeasurementTypesForDatasetAsync(sMyDatasetId, measurementTypeNames);

         // Create a property type to be associated with the measurement
         var propertyTypeNames = new List<PropertyType>
         {
            new PropertyType
            {
               Name = "DeviceMode",
               Format = PropertyTypeFormat.Text,
               Group = "My Properties"
            }
         };
         var createdPropertyTypes = await sClientApi.CreatePropertyTypesForDatasetAsync(sMyDatasetId, propertyTypeNames);

         // Upload a passing measurement using the above types
         var properties = new ObservableCollection<Property>
         {
            new Property
            {
               PropertyTypeId = createdPropertyTypes[0].Id,
               Value = "Mode A"
            }
         };
         var measurements = new List<Measurement>
         {
            new Measurement
            {
               MeasurementTypeId = createdMeasurementTypes[0].Id,
               Value = 1.5E+09,
               Properties = properties
            }
         };
         await sClientApi.CreateMeasurementsAsync(sMyDatasetId, measurements);
         Console.WriteLine("Posted a passing Frequency measurement to dataset.");
         // Upload a passing measurement using the above types
         properties = new ObservableCollection<Property>
         {
            new Property
            {
               PropertyTypeId = createdPropertyTypes[0].Id,
               Value = "Mode B"
            }
         };
         measurements = new List<Measurement>
         {
            new Measurement
            {
               MeasurementTypeId = createdMeasurementTypes[0].Id,
               Value = 0.5E+09,
               Properties = properties
            }
         };
         await sClientApi.CreateMeasurementsAsync(sMyDatasetId, measurements);
         Console.WriteLine("Posted a failing Frequency measurement to dataset.");
      }

      private static async Task ReportServerContents()
      {
         var lines = new List<string> {"URL: " + sUrl};

         foreach (var dataset in sDatasetsOnServer)
         {
            lines.Add(Environment.NewLine + "===============================================");
            lines.Add("Dataset: " + dataset.Name);

            lines.Add(string.Empty);
            var measurementTypes = await sClientApi.GetMeasurementTypesForDatasetAsync(sMyDatasetId);

            foreach (var measurementType in measurementTypes)
            {
               var entry = "Measurement Type: " + measurementType.Name;

               if (!string.IsNullOrEmpty(measurementType.Units))
               {
                  entry += ", Units = " + measurementType.Units;
               }

               if (measurementType.LowerLimit.HasValue)
               {
                  entry += ", LowerLimit = " + measurementType.LowerLimit;
               }

               if (measurementType.UpperLimit.HasValue)
               {
                  entry += ", UpperLimit = " + measurementType.UpperLimit;
               }

               lines.Add(entry);
            }

            lines.Add(string.Empty);
            var propertyTypes = await sClientApi.GetPropertyTypesAsync(sMyDatasetId);

            foreach (var propertyType in propertyTypes)
            {
               var entry = "Property Type: " + propertyType.Name + ", Format = " + propertyType.Format;

               if (!string.IsNullOrEmpty(propertyType.Group))
               {
                  entry += ", Group = " + propertyType.Group;
               }

               lines.Add(entry);
            }

            lines.Add(string.Empty);

            foreach (var measurementType in measurementTypes)
            {
               if (!measurementType.Id.HasValue)
               {
                  continue; // Corrupted
               }

               var measurements = await sClientApi.GetMeasurementsAsync(sMyDatasetId, measurementType.Id.Value);

               foreach (var measurement in measurements)
               {
                  if (!measurement.Value.HasValue)
                  {
                     continue; // Corrupted
                  }

                  lines.Add(measurementType.Name + " = " + measurement.Value.Value);
                  lines.Add("\tProperties:");

                  foreach (var property in measurement.Properties)
                  {
                     var propertyType = propertyTypes.FirstOrDefault(pt => pt.Id == property.PropertyTypeId);
                     var propertyName = propertyType == null ? "(unknown)" : propertyType.Name;
                     lines.Add("\t\t" + propertyName + " = " + property.Value);
                  }
               }
            }
         }

         var fullPath = Path.Combine(sOutputPath, "output.txt");
         File.WriteAllLines(fullPath, lines);
         Console.WriteLine("Server contents dumped to file: " + fullPath);
      }

      private static async Task LogOut()
      {
         if (sClientApi != null)
         {
            await sClientApi.LogoutAsync();
         }
      }

      private static void PressAnyKeyToClose()
      {
         Console.WriteLine(Environment.NewLine + "Press <enter> to close");
         Console.ReadLine();
         Environment.Exit(1);
      }
   }
}